/*
 * disk.c -- disk simulation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "curslib.h"
#include "error.h"
#include "arith.h"
#include "cpu.h"
#include "memory.h"
#include "timer.h"
#include "disk.h"


static bool debug = FALSE;

static FILE *diskImage;
static long numSectors;

static octa diskCtrl;
static octa diskCnt;
static octa diskSct;
static octa diskDma;
static octa diskCap;

static long lastSct;


static octa readOcta(byte *p) {
  octa data;

  data.h = ((tetra) *(p + 0)) << 24 |
           ((tetra) *(p + 1)) << 16 |
           ((tetra) *(p + 2)) <<  8 |
           ((tetra) *(p + 3)) <<  0;
  data.l = ((tetra) *(p + 4)) << 24 |
           ((tetra) *(p + 5)) << 16 |
           ((tetra) *(p + 6)) <<  8 |
           ((tetra) *(p + 7)) <<  0;
  return data;
}


static void writeOcta(byte *p, octa data) {
  *(p + 0) = (byte) (data.h >> 24);
  *(p + 1) = (byte) (data.h >> 16);
  *(p + 2) = (byte) (data.h >>  8);
  *(p + 3) = (byte) (data.h >>  0);
  *(p + 4) = (byte) (data.l >> 24);
  *(p + 5) = (byte) (data.l >> 16);
  *(p + 6) = (byte) (data.l >>  8);
  *(p + 7) = (byte) (data.l >>  0);
}


static void diskCallback(void) {
  byte buffer[SECTOR_SIZE];
  int i, j;
  octa data;

  if (debug) {
    cPrintf("\n**** DISK CALLBACK ****\n");
  }
  if (diskCap.l != 0 &&
      diskCnt.h == 0 && diskCnt.l != 0 &&
      diskSct.h == 0 && diskSct.l < diskCap.l &&
      diskSct.l + diskCnt.l <= diskCap.l) {
    /* do the transfer */
    if (fseek(diskImage, diskSct.l * SECTOR_SIZE, SEEK_SET) != 0) {
      error("cannot position to sector in disk image");
    }
    if (diskCtrl.l & DISK_WRT) {
      /* memory --> disk */
      for (i = 0; i < diskCnt.l; i++) {
        for (j = 0; j < SECTOR_SIZE; j += 8) {
          data = memoryRead64(diskDma);
          writeOcta(buffer + j, data);
          if (debug) {
            cPrintf("\n**** m[0x%08X%08X] --> 0x%08X%08X ****\n",
                    diskDma.h, diskDma.l, data.h, data.l);
          }
          diskDma = incrOcta(diskDma, 8);
        }
        if (fwrite(buffer, SECTOR_SIZE, 1, diskImage) != 1) {
          error("cannot write to disk image");
        }
        diskSct = incrOcta(diskSct, 1);
      }
    } else {
      /* disk --> memory */
      for (i = 0; i < diskCnt.l; i++) {
        if (fread(buffer, SECTOR_SIZE, 1, diskImage) != 1) {
          error("cannot read from disk image");
        }
        for (j = 0; j < SECTOR_SIZE; j += 8) {
          data = readOcta(buffer + j);
          memoryWrite64(diskDma, data);
          if (debug) {
            cPrintf("\n**** m[0x%08X%08X] <-- 0x%08X%08X ****\n",
                    diskDma.h, diskDma.l, data.h, data.l);
          }
          diskDma = incrOcta(diskDma, 8);
        }
        diskSct = incrOcta(diskSct, 1);
      }
    }
    lastSct = diskSct.l;
  } else {
    /* sectors requested exceed disk capacity */
    /* or there were zero sectors requested */
    /* or we have no disk at all */
    diskCtrl.l |= DISK_ERR;
  }
  diskCtrl.l &= ~DISK_BUSY;
  if (diskCtrl.l & DISK_IEN) {
    /* !!!!! raise disk interrupt !!!!! */
  }
}


octa diskRead(int devNo, tetra addr) {
  octa data;

  if (debug) {
    cPrintf("\n**** DISK %d READ from 0x%08X", devNo, addr);
  }
  if (devNo != 0) {
    /* currently at most one disk present */
    data.h = 0;
    data.l = 0;
  } else
  if (addr == DISK_CTRL) {
    data = diskCtrl;
  } else
  if (addr == DISK_CNT) {
    /* this register is write-only */
    data.h = 0;
    data.l = 0;
  } else
  if (addr == DISK_SCT) {
    /* this register is write-only */
    data.h = 0;
    data.l = 0;
  } else
  if (addr == DISK_DMA) {
    /* this register is write-only */
    data.h = 0;
    data.l = 0;
  } else
  if (addr == DISK_CAP) {
    data = diskCap;
  } else {
    /* illegal register */
    data.h = 0;
    data.l = 0;
  }
  if (debug) {
    cPrintf(", data = 0x%08X%08X ****\n", data.h, data.l);
  }
  return data;
}


void diskWrite(int devNo, tetra addr, octa data) {
  if (debug) {
    cPrintf("\n**** DISK %d WRITE to 0x%08X, data = 0x%08X%08X ****\n",
            devNo, addr, data.h, data.l);
  }
  if (devNo != 0) {
    /* currently at most one disk present */
  } else
  if (addr == DISK_CTRL) {
    if (data.l & DISK_WRT) {
      diskCtrl.l |= DISK_WRT;
    } else {
      diskCtrl.l &= ~DISK_WRT;
    }
    if (data.l & DISK_IEN) {
      diskCtrl.l |= DISK_IEN;
    } else {
      diskCtrl.l &= ~DISK_IEN;
    }
    if (data.l & DISK_STRT) {
      diskCtrl.l |= DISK_BUSY;
      diskCtrl.l &= ~DISK_ERR;
      /* only start a disk operation if disk is present */
      if (diskCap.l != 0) {
        timerStart(DISK_DELAY +
                   (abs(diskSct.l - lastSct) * DISK_SEEK) / diskCap.l,
                   diskCallback);
      }
    }
  } else
  if (addr == DISK_CNT) {
    diskCnt = data;
  } else
  if (addr == DISK_SCT) {
    diskSct = data;
  } else
  if (addr == DISK_DMA) {
    diskDma = data;
  } else
  if (addr == DISK_CAP) {
    /* this register is read-only */
  } else {
    /* illegal register */
  }
}


void diskReset(void) {
  cPrintf("Resetting Disk...\n");
  diskCtrl.h = 0;
  diskCtrl.l = 0;
  diskCnt.h = 0;
  diskCnt.l = 0;
  diskSct.h = 0;
  diskSct.l = 0;
  diskDma.h = 0;
  diskDma.l = 0;
  diskCap.h = 0;
  diskCap.l = numSectors;
  lastSct = 0;
}


void diskInit(char *diskImageName) {
  long numBytes;

  if (diskImageName == NULL) {
    /* do not install disk */
    diskImage = NULL;
    numSectors = 0;
  } else {
    /* try to install disk */
    diskImage = fopen(diskImageName, "r+b");
    if (diskImage == NULL) {
      error("cannot open disk image '%s'", diskImageName);
    }
    fseek(diskImage, 0, SEEK_END);
    numBytes = ftell(diskImage);
    fseek(diskImage, 0, SEEK_SET);
    numSectors = numBytes / SECTOR_SIZE;
    cPrintf("Disk of size %ld bytes (%ld sectors) installed.\n",
            numBytes, numSectors);
  }
  diskReset();
}


void diskExit(void) {
  if (diskImage == NULL) {
    /* disk not installed */
    return;
  }
  fclose(diskImage);
}
