/*
 * instr.c -- instruction encoding
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "curslib.h"
#include "error.h"
#include "instr.h"


/*
 * This is the GIMMIX machine instruction set.
 * The table below needs no particular order
 * and may have gaps in the instruction encoding.
 */
Instr instrTbl[] = {
  { "TRAP",    I_I8I8I8, 0x00 },
  { "MUL",     I_RRR,    0x18 },
  { "MULI",    I_RRI8,   0x19 },
  { "MULU",    I_RRR,    0x1A },
  { "MULUI",   I_RRI8,   0x1B },
  { "DIV",     I_RRR,    0x1C },
  { "DIVI",    I_RRI8,   0x1D },
  { "DIVU",    I_RRR,    0x1E },
  { "DIVUI",   I_RRI8,   0x1F },
  { "ADD",     I_RRR,    0x20 },
  { "ADDI",    I_RRI8,   0x21 },
  { "ADDU",    I_RRR,    0x22 },
  { "ADDUI",   I_RRI8,   0x23 },
  { "SUB",     I_RRR,    0x24 },
  { "SUBI",    I_RRI8,   0x25 },
  { "SUBU",    I_RRR,    0x26 },
  { "SUBUI",   I_RRI8,   0x27 },
  { "2ADDU",   I_RRR,    0x28 },
  { "2ADDUI",  I_RRI8,   0x29 },
  { "4ADDU",   I_RRR,    0x2A },
  { "4ADDUI",  I_RRI8,   0x2B },
  { "8ADDU",   I_RRR,    0x2C },
  { "8ADDUI",  I_RRI8,   0x2D },
  { "16ADDU",  I_RRR,    0x2E },
  { "16ADDUI", I_RRI8,   0x2F },
  { "CMP",     I_RRR,    0x30 },
  { "CMPI",    I_RRI8,   0x31 },
  { "CMPU",    I_RRR,    0x32 },
  { "CMPUI",   I_RRI8,   0x33 },
  { "NEG",     I_RI8R,   0x34 },
  { "NEGI",    I_RI8I8,  0x35 },
  { "NEGU",    I_RI8R,   0x36 },
  { "NEGUI",   I_RI8I8,  0x37 },
  { "SL",      I_RRR,    0x38 },
  { "SLI",     I_RRI8,   0x39 },
  { "SLU",     I_RRR,    0x3A },
  { "SLUI",    I_RRI8,   0x3B },
  { "SR",      I_RRR,    0x3C },
  { "SRI",     I_RRI8,   0x3D },
  { "SRU",     I_RRR,    0x3E },
  { "SRUI",    I_RRI8,   0x3F },
  { "BN",      I_RF16,   0x40 },
  { "BNB",     I_RB16,   0x41 },
  { "BZ",      I_RF16,   0x42 },
  { "BZB",     I_RB16,   0x43 },
  { "BP",      I_RF16,   0x44 },
  { "BPB",     I_RB16,   0x45 },
  { "BOD",     I_RF16,   0x46 },
  { "BODB",    I_RB16,   0x47 },
  { "BNN",     I_RF16,   0x48 },
  { "BNNB",    I_RB16,   0x49 },
  { "BNZ",     I_RF16,   0x4A },
  { "BNZB",    I_RB16,   0x4B },
  { "BNP",     I_RF16,   0x4C },
  { "BNPB",    I_RB16,   0x4D },
  { "BEV",     I_RF16,   0x4E },
  { "BEVB",    I_RB16,   0x4F },
  { "PBN",     I_RF16,   0x50 },
  { "PBNB",    I_RB16,   0x51 },
  { "PBZ",     I_RF16,   0x52 },
  { "PBZB",    I_RB16,   0x53 },
  { "PBP",     I_RF16,   0x54 },
  { "PBPB",    I_RB16,   0x55 },
  { "PBOD",    I_RF16,   0x56 },
  { "PBODB",   I_RB16,   0x57 },
  { "PBNN",    I_RF16,   0x58 },
  { "PBNNB",   I_RB16,   0x59 },
  { "PBNZ",    I_RF16,   0x5A },
  { "PBNZB",   I_RB16,   0x5B },
  { "PBNP",    I_RF16,   0x5C },
  { "PBNPB",   I_RB16,   0x5D },
  { "PBEV",    I_RF16,   0x5E },
  { "PBEVB",   I_RB16,   0x5F },
  { "CSN",     I_RRR,    0x60 },
  { "CSNI",    I_RRI8,   0x61 },
  { "CSZ",     I_RRR,    0x62 },
  { "CSZI",    I_RRI8,   0x63 },
  { "CSP",     I_RRR,    0x64 },
  { "CSPI",    I_RRI8,   0x65 },
  { "CSOD",    I_RRR,    0x66 },
  { "CSODI",   I_RRI8,   0x67 },
  { "CSNN",    I_RRR,    0x68 },
  { "CSNNI",   I_RRI8,   0x69 },
  { "CSNZ",    I_RRR,    0x6A },
  { "CSNZI",   I_RRI8,   0x6B },
  { "CSNP",    I_RRR,    0x6C },
  { "CSNPI",   I_RRI8,   0x6D },
  { "CSEV",    I_RRR,    0x6E },
  { "CSEVI",   I_RRI8,   0x6F },
  { "ZSN",     I_RRR,    0x70 },
  { "ZSNI",    I_RRI8,   0x71 },
  { "ZSZ",     I_RRR,    0x72 },
  { "ZSZI",    I_RRI8,   0x73 },
  { "ZSP",     I_RRR,    0x74 },
  { "ZSPI",    I_RRI8,   0x75 },
  { "ZSOD",    I_RRR,    0x76 },
  { "ZSODI",   I_RRI8,   0x77 },
  { "ZSNN",    I_RRR,    0x78 },
  { "ZSNNI",   I_RRI8,   0x79 },
  { "ZSNZ",    I_RRR,    0x7A },
  { "ZSNZI",   I_RRI8,   0x7B },
  { "ZSNP",    I_RRR,    0x7C },
  { "ZSNPI",   I_RRI8,   0x7D },
  { "ZSEV",    I_RRR,    0x7E },
  { "ZSEVI",   I_RRI8,   0x7F },
  { "LDB",     I_RRR,    0x80 },
  { "LDBI",    I_RRI8,   0x81 },
  { "LDBU",    I_RRR,    0x82 },
  { "LDBUI",   I_RRI8,   0x83 },
  { "LDW",     I_RRR,    0x84 },
  { "LDWI",    I_RRI8,   0x85 },
  { "LDWU",    I_RRR,    0x86 },
  { "LDWUI",   I_RRI8,   0x87 },
  { "LDT",     I_RRR,    0x88 },
  { "LDTI",    I_RRI8,   0x89 },
  { "LDTU",    I_RRR,    0x8A },
  { "LDTUI",   I_RRI8,   0x8B },
  { "LDO",     I_RRR,    0x8C },
  { "LDOI",    I_RRI8,   0x8D },
  { "LDOU",    I_RRR,    0x8E },
  { "LDOUI",   I_RRI8,   0x8F },
  { "LDHT",    I_RRR,    0x92 },
  { "LDHTI",   I_RRI8,   0x93 },
  { "CSWAP",   I_RRR,    0x94 },
  { "CSWAPI",  I_RRI8,   0x95 },
  { "LDUNC",   I_RRR,    0x96 },
  { "LDUNCI",  I_RRI8,   0x97 },
  { "PRELD",   I_I8RR,   0x9A },
  { "PRELDI",  I_I8RI8,  0x9B },
  { "PREGO",   I_I8RR,   0x9C },
  { "PREGOI",  I_I8RI8,  0x9D },
  { "GO",      I_RRR,    0x9E },
  { "GOI",     I_RRI8,   0x9F },
  { "STB",     I_RRR,    0xA0 },
  { "STBI",    I_RRI8,   0xA1 },
  { "STBU",    I_RRR,    0xA2 },
  { "STBUI",   I_RRI8,   0xA3 },
  { "STW",     I_RRR,    0xA4 },
  { "STWI",    I_RRI8,   0xA5 },
  { "STWU",    I_RRR,    0xA6 },
  { "STWUI",   I_RRI8,   0xA7 },
  { "STT",     I_RRR,    0xA8 },
  { "STTI",    I_RRI8,   0xA9 },
  { "STTU",    I_RRR,    0xAA },
  { "STTUI",   I_RRI8,   0xAB },
  { "STO",     I_RRR,    0xAC },
  { "STOI",    I_RRI8,   0xAD },
  { "STOU",    I_RRR,    0xAE },
  { "STOUI",   I_RRI8,   0xAF },
  { "STHT",    I_RRR,    0xB2 },
  { "STHTI",   I_RRI8,   0xB3 },
  { "STCO",    I_I8RR,   0xB4 },
  { "STCOI",   I_I8RI8,  0xB5 },
  { "STUNC",   I_RRR,    0xB6 },
  { "STUNCI",  I_RRI8,   0xB7 },
  { "PREST",   I_I8RR,   0xBA },
  { "PRESTI",  I_I8RI8,  0xBB },
  { "PUSHGO",  I_RRR,    0xBE },
  { "PUSHGOI", I_RRI8,   0xBF },
  { "OR",      I_RRR,    0xC0 },
  { "ORI",     I_RRI8,   0xC1 },
  { "ORN",     I_RRR,    0xC2 },
  { "ORNI",    I_RRI8,   0xC3 },
  { "NOR",     I_RRR,    0xC4 },
  { "NORI",    I_RRI8,   0xC5 },
  { "XOR",     I_RRR,    0xC6 },
  { "XORI",    I_RRI8,   0xC7 },
  { "AND",     I_RRR,    0xC8 },
  { "ANDI",    I_RRI8,   0xC9 },
  { "ANDN",    I_RRR,    0xCA },
  { "ANDNI",   I_RRI8,   0xCB },
  { "NAND",    I_RRR,    0xCC },
  { "NANDI",   I_RRI8,   0xCD },
  { "NXOR",    I_RRR,    0xCE },
  { "NXORI",   I_RRI8,   0xCF },
  { "SETH",    I_RI16,   0xE0 },
  { "SETMH",   I_RI16,   0xE1 },
  { "SETML",   I_RI16,   0xE2 },
  { "SETL",    I_RI16,   0xE3 },
  { "INCH",    I_RI16,   0xE4 },
  { "INCMH",   I_RI16,   0xE5 },
  { "INCML",   I_RI16,   0xE6 },
  { "INCL",    I_RI16,   0xE7 },
  { "ORH",     I_RI16,   0xE8 },
  { "ORMH",    I_RI16,   0xE9 },
  { "ORML",    I_RI16,   0xEA },
  { "ORL",     I_RI16,   0xEB },
  { "ANDNH",   I_RI16,   0xEC },
  { "ANDNMH",  I_RI16,   0xED },
  { "ANDNML",  I_RI16,   0xEE },
  { "ANDNL",   I_RI16,   0xEF },
  { "JMP",     I_F24,    0xF0 },
  { "JMPB",    I_B24,    0xF1 },
  { "PUSHJ",   I_RF16,   0xF2 },
  { "PUSHJB",  I_RB16,   0xF3 },
  { "GETA",    I_RF16,   0xF4 },
  { "GETAB",   I_RB16,   0xF5 },
  { "PUT",     I_S8Z8R,  0xF6 },
  { "PUTI",    I_S8Z8I8, 0xF7 },
  { "POP",     I_I8I16,  0xF8 },
  { "SWYM",    I_I8I8I8, 0xFD },
  { "GET",     I_RZ8S8,  0xFE },
  { "TRIP",    I_I8I8I8, 0xFF }
};


Instr *instrCodeTbl[256];


static int instrCompare(const void *instr1, const void *instr2) {
  return strcmp(((Instr *) instr1)->name, ((Instr *) instr2)->name);
}


void initInstrTable(void) {
  int i;

  /* first sort instruction table alphabetically */
  qsort(instrTbl, sizeof(instrTbl)/sizeof(instrTbl[0]),
        sizeof(instrTbl[0]), instrCompare);
  /* then initialize instruction code table */
  for (i = 0; i < 256; i++) {
    instrCodeTbl[i] = NULL;
  }
  for (i = 0; i < sizeof(instrTbl)/sizeof(instrTbl[0]); i++) {
    instrCodeTbl[instrTbl[i].opcode] = &instrTbl[i];
  }
}


Instr *lookupInstr(char *name) {
  int lo, hi, tst;
  int res;

  lo = 0;
  hi = sizeof(instrTbl) / sizeof(instrTbl[0]);
  while (lo <= hi) {
    tst = (lo + hi) / 2;
    res = strcmp(instrTbl[tst].name, name);
    if (res == 0) {
      return &instrTbl[tst];
    }
    if (res < 0) {
      lo = tst + 1;
    } else {
      hi = tst - 1;
    }
  }
  return NULL;
}
