/*
 * term.c -- terminal simulation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "curslib.h"
#include "error.h"
#include "cpu.h"
#include "timer.h"
#include "term.h"


static bool debug = FALSE;

static octa rcvrCtrl;
static octa rcvrData;

static octa xmtrCtrl;
static octa xmtrData;


static void rcvrCallback(void) {
  int c;

  if (debug) {
    cPrintf("\n**** TERM RCVR CALLBACK ****\n");
  }
  timerStart(TERM_RCVR_MSEC, rcvrCallback);
  c = cGetChar();
  if (c == -1) {
    /* no character typed */
    return;
  }
  if (c == TERM_SIM_ESC) {
    /* simulation escape character typed */
    cpuSetBreak(cpuGetPC());
  } else {
    /* any other character typed */
    rcvrData.l = c & 0xFF;
    rcvrCtrl.l |= TERM_RCVR_RDY;
    if (rcvrCtrl.l & TERM_RCVR_IEN) {
      /* !!!!! raise terminal rcvr interrupt !!!!! */
    }
  }
}


static void xmtrCallback(void) {
  if (debug) {
    cPrintf("\n**** TERM XMTR CALLBACK ****\n");
  }
  cPutChar(xmtrData.l);
  xmtrCtrl.l |= TERM_XMTR_RDY;
  if (xmtrCtrl.l & TERM_XMTR_IEN) {
    /* !!!!! raise terminal xmtr interrupt !!!!! */
  }
}


octa termRead(int devNo, tetra addr) {
  octa data;

  if (debug) {
    cPrintf("\n**** TERM %d READ from 0x%08X", devNo, addr);
  }
  if (devNo != 0) {
    /* currently only one terminal present */
    data.h = 0;
    data.l = 0;
  } else
  if (addr == TERM_RCVR_CTRL) {
    data = rcvrCtrl;
  } else
  if (addr == TERM_RCVR_DATA) {
    rcvrCtrl.l &= ~TERM_RCVR_RDY;
    data = rcvrData;
  } else
  if (addr == TERM_XMTR_CTRL) {
    data = xmtrCtrl;
  } else
  if (addr == TERM_XMTR_DATA) {
    /* this register is write-only */
    data.h = 0;
    data.l = 0;
  } else {
    /* illegal register */
    data.h = 0;
    data.l = 0;
  }
  if (debug) {
    cPrintf(", data = 0x%08X%08X ****\n", data.h, data.l);
  }
  return data;
}


void termWrite(int devNo, tetra addr, octa data) {
  if (debug) {
    cPrintf("\n**** TERM %d WRITE to 0x%08X, data = 0x%08X%08X ****\n",
            devNo, addr, data.h, data.l);
  }
  if (devNo != 0) {
    /* currently only one terminal present */
  } else
  if (addr == TERM_RCVR_CTRL) {
    if (data.l & TERM_RCVR_IEN) {
      rcvrCtrl.l |= TERM_RCVR_IEN;
    } else {
      rcvrCtrl.l &= ~TERM_RCVR_IEN;
    }
  } else
  if (addr == TERM_RCVR_DATA) {
    /* this register is read-only */
  } else
  if (addr == TERM_XMTR_CTRL) {
    if (data.l & TERM_XMTR_IEN) {
      xmtrCtrl.l |= TERM_XMTR_IEN;
    } else {
      xmtrCtrl.l &= ~TERM_XMTR_IEN;
    }
  } else
  if (addr == TERM_XMTR_DATA) {
    xmtrData.l = data.l & 0xFF;
    xmtrCtrl.l &= ~TERM_XMTR_RDY;
    timerStart(TERM_XMTR_MSEC, xmtrCallback);
  } else {
    /* illegal register */
  }
}


void termReset(void) {
  cPrintf("Resetting Terminal...\n");
  rcvrCtrl.h = 0;
  rcvrCtrl.l = 0;
  rcvrData.h = 0;
  rcvrData.l = 0;
  timerStart(TERM_RCVR_MSEC, rcvrCallback);
  xmtrCtrl.h = 0;
  xmtrCtrl.l = TERM_XMTR_RDY;
  xmtrData.h = 0;
  xmtrData.l = 0;
}


void termInit(void) {
  termReset();
}


void termExit(void) {
}
