/*
 * malloc.c -- allocate and free units from a map
 */


#include "../h/map.h"
#include "malloc.h"


/*
 * Allocate 'size' units from the given map.
 * Return the base of the allocated space.
 * In a map, the addresses are increasing and
 * the list is terminated by a 0 size.
 * The core map unit is 4K bytes, and the
 * swap map unit is also 4K bytes.
 * Algorithm is first-fit.
 */
unsigned int malloc(struct map *mp, unsigned int size) {
  struct map *bp;
  unsigned int a;

  for (bp = mp; bp->m_size != 0; bp++) {
    if (bp->m_size >= size) {
      a = bp->m_addr;
      bp->m_addr += size;
      bp->m_size -= size;
      if (bp->m_size == 0) {
        /* copy down rest of list */
        do {
          bp++;
          (bp-1)->m_addr = bp->m_addr;
          (bp-1)->m_size = bp->m_size;
        } while (bp->m_size != 0);
      }
      return a;
    }
  }
  return 0;
}


/*
 * Free the previously allocated space 'a'
 * of 'size' units into the specified map.
 * Sort 'a' into map and combine on one or
 * both ends if possible.
 */
void mfree(struct map *mp, unsigned int size, unsigned int a) {
  struct map *bp;
  unsigned int t;

  for (bp = mp; bp->m_addr <= a && bp->m_size != 0; bp++) ;
  if (bp != mp && (bp-1)->m_addr + (bp-1)->m_size == a) {
    /* merge with previous entry */
    (bp-1)->m_size += size;
    if (a + size == bp->m_addr) {
      /* merge with following entry */
      (bp-1)->m_size += bp->m_size;
      while (bp->m_size != 0) {
        bp++;
        (bp-1)->m_addr = bp->m_addr;
        (bp-1)->m_size = bp->m_size;
      }
    }
  } else
  if (a + size == bp->m_addr && bp->m_size != 0) {
    /* merge with following entry */
    bp->m_addr -= size;
    bp->m_size += size;
  } else
  if (size != 0) {
    /* insert new entry */
    do {
      t = bp->m_addr;
      bp->m_addr = a;
      a = t;
      t = bp->m_size;
      bp->m_size = size;
      size = t;
      bp++;
    } while (size != 0);
  }
}
